@@ -160,12 +160,12 @@ def _fft_helper(x, win, detrend_func, nperseg, noverlap, nfft, sides):
   if x.dtype.kind == 'i':
     x = x.astype(win.dtype)
 
+  *batch_shape, signal_length = x.shape
   # Created strided array of data segments
   if nperseg == 1 and noverlap == 0:
     result = x[..., np.newaxis]
   else:
     step = nperseg - noverlap
-    *batch_shape, signal_length = x.shape
     batch_shape = tuple(batch_shape)
     x = x.reshape((int(np.prod(batch_shape)), signal_length))[..., np.newaxis]
     result = jax.lax.conv_general_dilated_patches(
@@ -492,3 +492,138 @@ def welch(x, fs=1.0, window='hann', nperseg=None, noverlap=None, nfft=None,
                    axis=axis, average=average)
 
   return freqs, Pxx.real
+
+
+def _overlap_and_add(x, step_size):
+  """Utility function compatible with tf.signal.overlap_and_add.
+
+  Args:
+    x: An array with `(..., frames, frame_length)`-shape.
+    step_size: An integer denoting overlap offsets. Must be less than
+      `frame_length`.
+
+  Returns:
+    An array with `(..., output_size)`-shape containing overlapped signal.
+  """
+  _check_arraylike("_overlap_and_add", x)
+  step_size = jax.core.concrete_or_error(int, step_size,
+                                        "step_size for overlap_and_add")
+  if x.ndim < 2:
+    raise ValueError('Input must have (..., frames, frame_length) shape.')
+
+  *batch_shape, nframes, segment_len = x.shape
+  flat_batchsize = np.prod(batch_shape, dtype=np.int64)
+  x = x.reshape((flat_batchsize, nframes, segment_len))
+  output_size = step_size * (nframes - 1) + segment_len
+  nstep_per_segment = 1 + (segment_len - 1) // step_size
+
+  # Here, we use shorter notation for axes.
+  # B: batch_size, N: nframes, S: nstep_per_segment,
+  # T: segment_len divided by S
+
+  padded_segment_len = nstep_per_segment * step_size
+  x = jnp.pad(x, ((0, 0), (0, 0), (0, padded_segment_len - segment_len)))
+  x = x.reshape((flat_batchsize, nframes, nstep_per_segment, step_size))
+
+  # For obtaining shifted signals, this routine reinterprets flattened array
+  # with a shrinked axis.  With appropriate truncation/ padding, this operation
+  # pushes the last padded elements of the previous row to the head of the
+  # current row.
+  # See implementation of `overlap_and_add` in Tensorflow for details.
+  x = x.transpose((0, 2, 1, 3))  # x: (B, S, N, T)
+  x = jnp.pad(x, ((0, 0), (0, 0), (0, nframes), (0, 0)))  # x: (B, S, N*2, T)
+  shrinked = x.shape[2] - 1
+  x = x.reshape((flat_batchsize, -1))
+  x = x[:, :(nstep_per_segment * shrinked * step_size)]
+  x = x.reshape((flat_batchsize, nstep_per_segment, shrinked * step_size))
+
+  # Finally, sum shifted segments, and truncate results to the output_size.
+  x = x.sum(axis=1)[:, :output_size]
+  return x.reshape(tuple(batch_shape) + (-1,))
+
+
+@_wraps(osp_signal.istft)
+def istft(Zxx, fs=1.0, window='hann', nperseg=None, noverlap=None, nfft=None,
+          input_onesided=True, boundary=True, time_axis=-1, freq_axis=-2):
+  # Input validation
+  _check_arraylike("istft", Zxx)
+  if Zxx.ndim < 2:
+    raise ValueError('Input stft must be at least 2d!')
+  freq_axis = canonicalize_axis(freq_axis, Zxx.ndim)
+  time_axis = canonicalize_axis(time_axis, Zxx.ndim)
+  if freq_axis == time_axis:
+    raise ValueError('Must specify differing time and frequency axes!')
+
+  Zxx = jnp.asarray(Zxx, dtype=jax.dtypes.canonicalize_dtype(
+      np.result_type(Zxx, np.complex64)))
+
+  n_default = (2 * (Zxx.shape[freq_axis] - 1) if input_onesided
+               else Zxx.shape[freq_axis])
+
+  nperseg = jax.core.concrete_or_error(int, nperseg or n_default,
+                                       "nperseg: segment length of STFT")
+  if nperseg < 1:
+    raise ValueError('nperseg must be a positive integer')
+
+  if nfft is None:
+    nfft = n_default
+    if input_onesided and nperseg == n_default + 1:
+      nfft += 1  # Odd nperseg, no FFT padding
+  else:
+    nfft = jax.core.concrete_or_error(int, nfft, "nfft of STFT")
+  if nfft < nperseg:
+    raise ValueError(
+        f'FFT length ({nfft}) must be longer than nperseg ({nperseg}).')
+
+  noverlap = jax.core.concrete_or_error(int, noverlap or nperseg // 2,
+                                        "noverlap of STFT")
+  if noverlap >= nperseg:
+    raise ValueError('noverlap must be less than nperseg.')
+  nstep = nperseg - noverlap
+
+  # Rearrange axes if necessary
+  if time_axis != Zxx.ndim - 1 or freq_axis != Zxx.ndim - 2:
+    outer_idxs = tuple(
+        idx for idx in range(Zxx.ndim) if idx not in {time_axis, freq_axis})
+    Zxx = jnp.transpose(Zxx, outer_idxs + (freq_axis, time_axis))
+
+  # Perform IFFT
+  ifunc = jax.numpy.fft.irfft if input_onesided else jax.numpy.fft.ifft
+  # xsubs: [..., T, N], N is the number of frames, T is the frame length.
+  xsubs = ifunc(Zxx, axis=-2, n=nfft)[..., :nperseg, :]
+
+  # Get window as array
+  if isinstance(window, (str, tuple)):
+    win = osp_signal.get_window(window, nperseg)
+    win = jnp.asarray(win)
+  else:
+    win = jnp.asarray(window)
+    if len(win.shape) != 1:
+      raise ValueError('window must be 1-D')
+    if win.shape[0] != nperseg:
+      raise ValueError('window must have length of {0}'.format(nperseg))
+  win = win.astype(xsubs.dtype)
+
+  xsubs *= win.sum()  # This takes care of the 'spectrum' scaling
+
+  # make win broadcastable over xsubs
+  win = win.reshape((1, ) * (xsubs.ndim - 2) + win.shape + (1,))
+  x = _overlap_and_add((xsubs * win).swapaxes(-2, -1), nstep)
+  win_squared = jnp.repeat((win * win), xsubs.shape[-1], axis=-1)
+  norm = _overlap_and_add(win_squared.swapaxes(-2, -1), nstep)
+
+  # Remove extension points
+  if boundary:
+    x = x[..., nperseg//2:-(nperseg//2)]
+    norm = norm[..., nperseg//2:-(nperseg//2)]
+  x /= jnp.where(norm > 1e-10, norm, 1.0)
+
+  # Put axes back
+  if x.ndim > 1:
+    if time_axis != Zxx.ndim - 1:
+      if freq_axis < time_axis:
+        time_axis -= 1
+      x = jnp.moveaxis(x, -1, time_axis)
+
+  time = jnp.arange(x.shape[0]) / fs
+  return time, x
