@@ -12,6 +12,8 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+# flake8: noqa: F401
+
 from contextlib import contextmanager
 import inspect
 import functools
@@ -19,7 +21,7 @@ from functools import partial
 import re
 import os
 import textwrap
-from typing import Dict, List, Generator, Sequence, Tuple, Union
+from typing import Dict, List, Generator, Sequence, Tuple, Union, NamedTuple
 import unittest
 import warnings
 import zlib
@@ -30,19 +32,28 @@ from absl.testing import parameterized
 import numpy as np
 import numpy.random as npr
 
+from jax import stages
 from jax._src import api
 from jax import core
 from jax._src import dtypes as _dtypes
 from jax import lax
 from jax._src.config import flags, bool_env, config
 from jax._src.util import prod, unzip2
-from jax.tree_util import tree_multimap, tree_all, tree_map, tree_reduce
+from jax.tree_util import tree_map, tree_all
 from jax._src.lib import xla_bridge
 from jax._src import dispatch
+from jax._src.public_test_util import (
+    _assert_numpy_allclose, _check_dtypes_match, _default_tolerance, _dtype, check_close, check_grads,
+    check_jvp, check_vjp, default_gradient_tolerance, default_tolerance, device_under_test, tolerance)
 from jax.interpreters import mlir
 from jax.interpreters import xla
 from jax.experimental.maps import Mesh
+from jax.interpreters.pxla import PartitionSpec
+from jax.experimental import pjit
 
+# This submodule includes private test utilities that are not exported to
+# jax.test_util. Functionality appearing here is for internal use only, and
+# may be changed or removed at any time and without any deprecation cycle.
 
 FLAGS = flags.FLAGS
 flags.DEFINE_string(
@@ -81,16 +92,6 @@ flags.DEFINE_string(
   'on the test name. If empty or unspecified, run all tests.'
 )
 
-EPS = 1e-4
-
-def _dtype(x):
-  if hasattr(x, 'dtype'):
-    return x.dtype
-  elif type(x) in _dtypes.python_scalar_dtypes:
-    return np.dtype(_dtypes.python_scalar_dtypes[type(x)])
-  else:
-    return np.asarray(x).dtype
-
 def num_float_bits(dtype):
   return _dtypes.finfo(_dtypes.canonicalize_dtype(dtype)).bits
 
@@ -136,64 +137,6 @@ def is_sequence(x):
   else:
     return True
 
-_default_tolerance = {
-  _dtypes.float0: 0,
-  np.dtype(np.bool_): 0,
-  np.dtype(np.int8): 0,
-  np.dtype(np.int16): 0,
-  np.dtype(np.int32): 0,
-  np.dtype(np.int64): 0,
-  np.dtype(np.uint8): 0,
-  np.dtype(np.uint16): 0,
-  np.dtype(np.uint32): 0,
-  np.dtype(np.uint64): 0,
-  np.dtype(_dtypes.bfloat16): 1e-2,
-  np.dtype(np.float16): 1e-3,
-  np.dtype(np.float32): 1e-6,
-  np.dtype(np.float64): 1e-15,
-  np.dtype(np.complex64): 1e-6,
-  np.dtype(np.complex128): 1e-15,
-}
-
-def default_tolerance():
-  if device_under_test() != "tpu":
-    return _default_tolerance
-  tol = _default_tolerance.copy()
-  tol[np.dtype(np.float32)] = 1e-3
-  tol[np.dtype(np.complex64)] = 1e-3
-  return tol
-
-default_gradient_tolerance = {
-  np.dtype(_dtypes.bfloat16): 1e-1,
-  np.dtype(np.float16): 1e-2,
-  np.dtype(np.float32): 2e-3,
-  np.dtype(np.float64): 1e-5,
-  np.dtype(np.complex64): 1e-3,
-  np.dtype(np.complex128): 1e-5,
-}
-
-def _assert_numpy_allclose(a, b, atol=None, rtol=None, err_msg=''):
-  if a.dtype == b.dtype == _dtypes.float0:
-    np.testing.assert_array_equal(a, b, err_msg=err_msg)
-    return
-  a = a.astype(np.float32) if a.dtype == _dtypes.bfloat16 else a
-  b = b.astype(np.float32) if b.dtype == _dtypes.bfloat16 else b
-  kw = {}
-  if atol: kw["atol"] = atol
-  if rtol: kw["rtol"] = rtol
-  with np.errstate(invalid='ignore'):
-    # TODO(phawkins): surprisingly, assert_allclose sometimes reports invalid
-    # value errors. It should not do that.
-    np.testing.assert_allclose(a, b, **kw, err_msg=err_msg)
-
-def tolerance(dtype, tol=None):
-  tol = {} if tol is None else tol
-  if not isinstance(tol, dict):
-    return tol
-  tol = {np.dtype(key): value for key, value in tol.items()}
-  dtype = _dtypes.canonicalize_dtype(np.dtype(dtype))
-  return tol.get(dtype, default_tolerance()[dtype])
-
 def _normalize_tolerance(tol):
   tol = tol or 0
   if isinstance(tol, dict):
@@ -209,167 +152,10 @@ def join_tolerance(tol1, tol2):
     out[k] = max(v, tol1.get(k, 0))
   return out
 
-def _assert_numpy_close(a, b, atol=None, rtol=None, err_msg=''):
-  a, b = np.asarray(a), np.asarray(b)
-  assert a.shape == b.shape
-  atol = max(tolerance(a.dtype, atol), tolerance(b.dtype, atol))
-  rtol = max(tolerance(a.dtype, rtol), tolerance(b.dtype, rtol))
-  _assert_numpy_allclose(a, b, atol=atol * a.size, rtol=rtol * b.size,
-                         err_msg=err_msg)
 
 def check_eq(xs, ys, err_msg=''):
   assert_close = partial(_assert_numpy_allclose, err_msg=err_msg)
-  tree_all(tree_multimap(assert_close, xs, ys))
-
-def check_close(xs, ys, atol=None, rtol=None, err_msg=''):
-  assert_close = partial(_assert_numpy_close, atol=atol, rtol=rtol,
-                         err_msg=err_msg)
-  tree_all(tree_multimap(assert_close, xs, ys))
-
-def _check_dtypes_match(xs, ys):
-  def _assert_dtypes_match(x, y):
-    if config.x64_enabled:
-      assert _dtype(x) == _dtype(y)
-    else:
-      assert (_dtypes.canonicalize_dtype(_dtype(x)) ==
-              _dtypes.canonicalize_dtype(_dtype(y)))
-  tree_all(tree_multimap(_assert_dtypes_match, xs, ys))
-
-
-def inner_prod(xs, ys):
-  def contract(x, y):
-    return np.real(np.dot(np.conj(x).reshape(-1), y.reshape(-1)))
-  return tree_reduce(np.add, tree_multimap(contract, xs, ys))
-
-
-def _safe_subtract(x, y, *, dtype):
-  """Subtraction that with `inf - inf == 0` semantics."""
-  with np.errstate(invalid='ignore'):
-    return np.where(np.equal(x, y), np.array(0, dtype),
-                    np.subtract(x, y, dtype=dtype))
-
-add = partial(tree_multimap, lambda x, y: np.add(x, y, dtype=_dtype(x)))
-sub = partial(tree_multimap, lambda x, y: np.subtract(x, y, dtype=_dtype(x)))
-safe_sub = partial(tree_multimap,
-                   lambda x, y: _safe_subtract(x, y, dtype=_dtype(x)))
-conj = partial(tree_map, lambda x: np.conj(x, dtype=_dtype(x)))
-
-def scalar_mul(xs, a):
-  def mul(x):
-    dtype = _dtype(x)
-    return np.multiply(x, np.array(a, dtype=dtype), dtype=dtype)
-  return tree_map(mul, xs)
-
-
-def rand_like(rng, x):
-  shape = np.shape(x)
-  dtype = _dtype(x)
-  randn = lambda: np.asarray(rng.randn(*shape), dtype=dtype)
-  if _dtypes.issubdtype(dtype, np.complexfloating):
-    return randn() + dtype.type(1.0j) * randn()
-  else:
-    return randn()
-
-
-def numerical_jvp(f, primals, tangents, eps=EPS):
-  delta = scalar_mul(tangents, eps)
-  f_pos = f(*add(primals, delta))
-  f_neg = f(*sub(primals, delta))
-  return scalar_mul(safe_sub(f_pos, f_neg), 0.5 / eps)
-
-
-def _merge_tolerance(tol, default):
-  if tol is None:
-    return default
-  if not isinstance(tol, dict):
-    return tol
-  out = default.copy()
-  for k, v in tol.items():
-    out[np.dtype(k)] = v
-  return out
-
-
-def check_jvp(f, f_jvp, args, atol=None, rtol=None, eps=EPS, err_msg=''):
-  atol = _merge_tolerance(atol, default_gradient_tolerance)
-  rtol = _merge_tolerance(rtol, default_gradient_tolerance)
-  rng = np.random.RandomState(0)
-  tangent = tree_map(partial(rand_like, rng), args)
-  v_out, t_out = f_jvp(args, tangent)
-  _check_dtypes_match(v_out, t_out)
-  v_out_expected = f(*args)
-  _check_dtypes_match(v_out, v_out_expected)
-  t_out_expected = numerical_jvp(f, args, tangent, eps=eps)
-  # In principle we should expect exact equality of v_out and v_out_expected,
-  # but due to nondeterminism especially on GPU (e.g., due to convolution
-  # autotuning) we only require "close".
-  check_close(v_out, v_out_expected, atol=atol, rtol=rtol,
-              err_msg=f'{err_msg} primal' if err_msg else 'primal')
-  check_close(t_out, t_out_expected, atol=atol, rtol=rtol,
-              err_msg=f'{err_msg} tangent' if err_msg else 'tangent')
-
-
-def check_vjp(f, f_vjp, args, atol=None, rtol=None, eps=EPS, err_msg=''):
-  atol = _merge_tolerance(atol, default_gradient_tolerance)
-  rtol = _merge_tolerance(rtol, default_gradient_tolerance)
-  _rand_like = partial(rand_like, np.random.RandomState(0))
-  v_out, vjpfun = f_vjp(*args)
-  v_out_expected = f(*args)
-  check_close(v_out, v_out_expected, atol=atol, rtol=rtol,
-              err_msg=f'{err_msg} primal' if err_msg else 'primal')
-  tangent = tree_map(_rand_like, args)
-  tangent_out = numerical_jvp(f, args, tangent, eps=eps)
-  cotangent = tree_map(_rand_like, v_out)
-  cotangent_out = conj(vjpfun(conj(cotangent)))
-  ip = inner_prod(tangent, cotangent_out)
-  ip_expected = inner_prod(tangent_out, cotangent)
-  check_close(ip, ip_expected, atol=atol, rtol=rtol,
-              err_msg=(f'{err_msg} cotangent projection'
-                       if err_msg else 'cotangent projection'))
-
-
-def check_grads(f, args, order,
-                modes=("fwd", "rev"), atol=None, rtol=None, eps=None):
-  """Check gradients from automatic differentiation against finite differences.
-
-  Gradients are only checked in a single randomly chosen direction, which
-  ensures that the finite difference calculation does not become prohibitively
-  expensive even for large input/output spaces.
-
-  Args:
-    f: function to check at ``f(*args)``.
-    args: tuple of argument values.
-    order: forward and backwards gradients up to this order are checked.
-    modes: lists of gradient modes to check ('fwd' and/or 'rev').
-    atol: absolute tolerance for gradient equality.
-    rtol: relative tolerance for gradient equality.
-    eps: step size used for finite differences.
-
-  Raises:
-    AssertionError: if gradients do not match.
-  """
-  args = tuple(args)
-  eps = eps or EPS
-
-  _check_jvp = partial(check_jvp, atol=atol, rtol=rtol, eps=eps)
-  _check_vjp = partial(check_vjp, atol=atol, rtol=rtol, eps=eps)
-
-  def _check_grads(f, args, order, err_msg=''):
-    if "fwd" in modes:
-      fwd_msg = f'JVP of {err_msg}' if err_msg else 'JVP'
-      _check_jvp(f, partial(api.jvp, f), args, err_msg=fwd_msg)
-      if order > 1:
-        _check_grads(partial(api.jvp, f), (args, args), order - 1, fwd_msg)
-
-    if "rev" in modes:
-      rev_msg = f'VJP of {err_msg}' if err_msg else 'VJP'
-      _check_vjp(f, partial(api.vjp, f), args, err_msg=rev_msg)
-      if order > 1:
-        def f_vjp(*args):
-          out_primal_py, vjp_py = api.vjp(f, *args)
-          return vjp_py(out_primal_py)
-        _check_grads(f_vjp, args, order - 1, rev_msg)
-
-  _check_grads(f, args, order)
+  tree_all(tree_map(assert_close, xs, ys))
 
 
 @contextmanager
@@ -432,9 +218,6 @@ def assert_num_jit_and_pmap_compilations(times):
     raise AssertionError(f"Expected exactly {times} XLA compilations, "
                          f"but executed {count[0]}")
 
-def device_under_test():
-  return FLAGS.jax_test_dut or xla_bridge.get_backend().platform
-
 def if_device_under_test(device_type: Union[str, Sequence[str]],
                          if_true, if_false):
   """Chooses `if_true` of `if_false` based on device_under_test."""
@@ -1165,6 +948,22 @@ class _cached_property:
     return self._value
 
 
+class _XLAShardingInfo(NamedTuple):
+  in_pspec: Tuple[PartitionSpec]
+  out_pspec: Tuple[PartitionSpec]
+  compiled: stages.Compiled
+
+
+def compile_and_get_sharding(pjitted_fn, mesh, global_inputs):
+  # TODO(yashkatariya): Check if the pjitted_fn comes from pjit.
+  inputs = [core.ShapedArray(i.shape, i.dtype) for i in global_inputs]
+  compiled = pjitted_fn.lower(*inputs, _global_avals=True).compile()
+  in_sharding, out_sharding = pjit._get_sharding_from_executable(
+      compiled.runtime_executable(), mesh)
+  return _XLAShardingInfo(in_pspec=in_sharding, out_pspec=out_sharding,
+                          compiled=compiled)
+
+
 class _LazyDtypes:
   """A class that unifies lists of supported dtypes.
 
@@ -1226,3 +1025,35 @@ class _LazyDtypes:
 
 
 dtypes = _LazyDtypes()
+
+
+class DeprecatedJaxTestCase(JaxTestCase):
+  def __init__(self, *args, **kwargs):
+    warnings.warn(textwrap.dedent("""\
+      jax.test_util.JaxTestCase is deprecated as of jax version 0.3.1:
+      The suggested replacement is to use parametrized.TestCase directly.
+      For tests that rely on custom asserts such as JaxTestCase.assertAllClose(),
+      the suggested replacement is to use standard numpy testing utilities such
+      as np.testing.assert_allclose(), which work directly with JAX arrays."""),
+      category=DeprecationWarning)
+    super().__init__(*args, **kwargs)
+
+
+class DeprecatedJaxTestLoader(JaxTestLoader):
+  def __init__(self, *args, **kwargs):
+    warnings.warn(
+      "jax.test_util.JaxTestLoader is deprecated as of jax version 0.3.1. Use absltest.TestLoader directly.",
+      category=DeprecationWarning)
+    super().__init__(*args, **kwargs)
+
+
+class DeprecatedBufferDonationTestCase(BufferDonationTestCase):
+  def __init__(self, *args, **kwargs):
+    warnings.warn(textwrap.dedent("""\
+      jax.test_util.JaxTestCase is deprecated as of jax version 0.3.1:
+      The suggested replacement is to use parametrized.TestCase directly.
+      For tests that rely on custom asserts such as JaxTestCase.assertAllClose(),
+      the suggested replacement is to use standard numpy testing utilities such
+      as np.testing.assert_allclose(), which work directly with JAX arrays."""),
+      category=DeprecationWarning)
+    super().__init__(*args, **kwargs)
