@@ -54,14 +54,21 @@ map, unsafe_map = safe_map, map
 
 # -------------------- jaxprs --------------------
 
+Effect = Any
+Effects = Set[Effect]
+
+no_effects: Effects = set()
+
 class Jaxpr:
   constvars: List[Var]
   invars: List[Var]
   outvars: List[Atom]
   eqns: List[JaxprEqn]
+  effects: Effects
 
   def __init__(self, constvars: Sequence[Var], invars: Sequence[Var],
-               outvars: Sequence[Atom], eqns: Sequence[JaxprEqn]):
+               outvars: Sequence[Atom], eqns: Sequence[JaxprEqn],
+               effects: Effects = no_effects):
     """
     Args:
       constvars: list of variables introduced for constants. Array constants are
@@ -70,11 +77,14 @@ class Jaxpr:
         the inputs to the Jaxpr.
       outvars: list of output variables.
       eqns: list of equations.
+      effects: set of effects. The effects on a jaxpr are a superset of the
+        union of the effects for each equation.
     """
     self.constvars = list(constvars)
     self.invars = list(invars)
     self.outvars = list(outvars)
     self.eqns = list(eqns)
+    self.effects = effects
 
   def __str__(self):
     return str(pp_jaxpr(self, JaxprPpContext(), JaxprPpSettings()))
@@ -92,6 +102,18 @@ class Jaxpr:
   def _repr_pretty_(self, p, cycle):
     return p.text(self.pretty_print(use_color=True))
 
+  def replace(self, *, constvars=None, invars=None, outvars=None, eqns=None,
+              effects=None):
+    constvars = self.constvars if constvars is None else constvars
+    invars =  self.invars if invars is None else invars
+    outvars = self.outvars if outvars is None else outvars
+    eqns = self.eqns if eqns is None else eqns
+    effects = self.effects if effects is None else effects
+    return Jaxpr(constvars=constvars, invars=invars, outvars=outvars, eqns=eqns,
+                 effects=effects)
+
+def join_effects(*effects: Effect) -> Effects:
+  return set.union(*effects) if effects else no_effects
 
 def jaxprs_in_params(params) -> Iterator[Jaxpr]:
   for val in params.values():
@@ -137,9 +159,18 @@ class ClosedJaxpr:
   def eqns(self):
     return self.jaxpr.eqns
 
+  @property
+  def effects(self) -> Effects:
+    return self.jaxpr.effects
+
   def map_jaxpr(self, f):
     return ClosedJaxpr(f(self.jaxpr), self.consts)
 
+  def replace(self, *, jaxpr=None, consts=None):
+    jaxpr = self.jaxpr if jaxpr is None else jaxpr
+    consts = self.consts if consts is None else consts
+    return ClosedJaxpr(jaxpr, consts)
+
   def __str__(self): return str(self.jaxpr)
   def __repr__(self): return repr(self.jaxpr)
 
@@ -164,16 +195,20 @@ class JaxprEqn(NamedTuple):
   outvars: List[Var]
   primitive: Primitive
   params: Dict[str, Any]
+  effects: Effects
   source_info: source_info_util.SourceInfo
 
   def __repr__(self):
     return str(pp_eqn(self, JaxprPpContext(), JaxprPpSettings())).rstrip()
 
-def new_jaxpr_eqn(invars, outvars, primitive, params, source_info=None):
+  def replace(self, *args, **kwargs):
+    return self._replace(*args, **kwargs)
+
+def new_jaxpr_eqn(invars, outvars, primitive, params, effects, source_info=None):
   if primitive.call_primitive:
     assert len(outvars) == len(params["call_jaxpr"].outvars)
   source_info = source_info or source_info_util.new_source_info()
-  return JaxprEqn(invars, outvars, primitive, params, source_info)
+  return JaxprEqn(invars, outvars, primitive, params, effects, source_info)
 
 
 @total_ordering
@@ -296,9 +331,13 @@ class Primitive:
     return impl
 
   def def_abstract_eval(self, abstract_eval):
-    self.abstract_eval = abstract_eval
+    self.abstract_eval = _effect_free_abstract_eval(abstract_eval)  # type: ignore[assignment]
     return abstract_eval
 
+  def def_effectful_abstract_eval(self, effectful_abstract_eval):
+    self.abstract_eval = effectful_abstract_eval  # type: ignore[assignment]
+    return effectful_abstract_eval
+
   def def_custom_bind(self, bind):
     self.bind = bind
     return bind
@@ -315,6 +354,11 @@ class Primitive:
     return [], params
 
 
+def _effect_free_abstract_eval(abstract_eval):
+  def abstract_eval_(*args, **kwargs):
+    return abstract_eval(*args, **kwargs), no_effects
+  return abstract_eval_
+
 # -------------------- lifting --------------------
 
 # TODO(mattjj): replace this approach with a primitive-keyed table of rules
@@ -990,6 +1034,18 @@ class AbstractUnit(AbstractValue):
 
 abstract_unit = AbstractUnit()
 
+class AbstractBInt(AbstractValue):
+  __slots__ = ['bound']
+  bound: int
+  def __init__(self, bound):
+    self.bound = bound
+  def str_short(self, short_dtypes=False) -> str:
+    return f'bint{{≤{self.bound}}}[]'
+  def __eq__(self, other):
+    return type(other) is AbstractBInt and self.bound == other.bound
+  def __hash__(self) -> int:
+    return hash((type(self), self.bound))
+
 def lattice_join(x: Optional[AbstractValue],
                  y: Optional[AbstractValue]) -> AbstractValue:
   if x is None:
@@ -1168,7 +1224,7 @@ AxisSize = Union[AxisSizeForTracing, AxisSizeForJaxprType,
 class DShapedArray(UnshapedArray):
   __slots__ = ['shape']
   shape: Tuple[AxisSize, ...]  # noqa: F821
-  array_abstraction_level = 2
+  array_abstraction_level: int = 2
 
   def __init__(self, shape, dtype, weak_type):
     self.shape = shape
@@ -1185,11 +1241,6 @@ class DShapedArray(UnshapedArray):
     return f'{dtype}[{shape}]'
   __str__ = __repr__ = str_short
 
-  def __eq__(self, other):
-    return (type(self) is type(other) and
-            self.dtype == other.dtype and self.shape == other.shape and
-            self.weak_type == other.weak_type)
-
   def update(self, shape=None, dtype=None, weak_type=None):
     if shape is None:
       shape = self.shape
@@ -1199,6 +1250,14 @@ class DShapedArray(UnshapedArray):
       weak_type = self.weak_type
     return DShapedArray(shape, dtype, weak_type)
 
+  def __eq__(self, other):
+    return (type(self) is type(other)
+            and self.dtype == other.dtype and self.shape == other.shape
+            and self.weak_type == other.weak_type)
+
+  def __hash__(self):
+    return hash((self.shape, self.dtype, self.weak_type))
+
 del AxisSize, AxisSizeForTracing, AxisSizeForJaxprType, \
     AxisSizeForJaxprTracingSpec
 
@@ -1371,6 +1430,7 @@ def raise_to_shaped(aval: AbstractValue, weak_type=None):
 
 raise_to_shaped_mappings : Dict[type, Callable] = {
   AbstractUnit: lambda aval, _: aval,
+  AbstractBInt: lambda aval, _: aval,
   AbstractToken: lambda aval, _: aval,
   Bot: lambda aval, _: aval,
   UnshapedArray: lambda aval, _: aval,
@@ -1714,14 +1774,17 @@ class CallPrimitive(Primitive):
     new_params = dict(params)
     jaxpr = new_params.pop('call_jaxpr')
     subfun = lu.hashable_partial(lu.wrap_init(eval_jaxpr), jaxpr, ())
+    if config.jax_dynamic_shapes:
+      subfun = lu.annotate(subfun, tuple((v.aval, True) for v in jaxpr.invars))
     return [subfun], new_params
 
 def call_bind(primitive: CallPrimitive, fun, *args, **params):
   top_trace = find_top_trace(args)
-  fun, env_trace_todo = process_env_traces_call(
+  fun_, env_trace_todo = process_env_traces_call(
       fun, primitive, top_trace and top_trace.level, tuple(params.items()))
   tracers = map(top_trace.full_raise, args)
-  outs = top_trace.process_call(primitive, fun, tracers, params)
+  fun_ = lu.annotate(fun_, fun.in_type)
+  outs = top_trace.process_call(primitive, fun_, tracers, params)
   return map(full_lower, apply_todos(env_trace_todo(), outs))
 
 @lu.transformation_with_aux
@@ -1887,13 +1950,28 @@ def _unmap_shaped_array(size: int, axis_name, axis: Optional[int],
   named_shape = dict(aval.named_shape)
   # TODO: Make this mandatory
   named_shape.pop(axis_name, None)
-  if axis is None: return aval.replace(named_shape=named_shape)
+  if axis is None: return aval.update(named_shape=named_shape)
   return ShapedArray(tuple_insert(aval.shape, axis, size), aval.dtype,
                      named_shape=named_shape, weak_type=aval.weak_type)
 
+def _map_dshaped_array(size: Union[int, Tracer], axis: Optional[int],
+                       aval: ShapedArray) -> ShapedArray:
+  assert False  # TODO(mattjj, dougalm)
+
+def _unmap_dshaped_array(
+    size: Union[int, Tracer], axis_name, axis: Optional[int],
+    aval: DShapedArray) -> DShapedArray:
+  if isinstance(size, int):
+    if axis is None: return aval
+    return DShapedArray(tuple_insert(aval.shape, axis, size), aval.dtype,
+                        weak_type=aval.weak_type)
+  else:
+    assert False  # TODO(mattjj, dougalm)
+
 AvalMapHandlerPair = Tuple[Callable, Callable]
 aval_mapping_handlers: Dict[Type, AvalMapHandlerPair] = {
     AbstractUnit: (_map_unit, _map_unit),
+    DShapedArray:   (_map_dshaped_array, _unmap_dshaped_array),
     ShapedArray:   (_map_shaped_array, _unmap_shaped_array),
     ConcreteArray: (_map_shaped_array, _unmap_shaped_array),
 }
@@ -2018,7 +2096,7 @@ def subst_axis_names_eqn(eqn: JaxprEqn, subst: AxisSubst, var_map: Dict[Var, Var
     e.eqn = eqn
     raise
   params = subst_axis_names(eqn.primitive, eqn.params, subst)
-  return new_jaxpr_eqn(invars, outvars, eqn.primitive, params, eqn.source_info)
+  return eqn.replace(invars=invars, outvars=outvars, params=params)
 
 def do_subst_axis_names_jaxpr(jaxpr: Union[Jaxpr, ClosedJaxpr], subst: AxisSubst):
   consts = None
@@ -2030,7 +2108,7 @@ def do_subst_axis_names_jaxpr(jaxpr: Union[Jaxpr, ClosedJaxpr], subst: AxisSubst
   constvars = [subst_axis_names_var(v, subst, var_map) for v in jaxpr.constvars]
   eqns = [subst_axis_names_eqn(eqn, subst, var_map) for eqn in jaxpr.eqns]
   outvars: List[Atom] = [v if isinstance(v, Literal) else var_map[v] for v in jaxpr.outvars]
-  new_jaxpr = Jaxpr(constvars, invars, outvars, eqns)
+  new_jaxpr = Jaxpr(constvars, invars, outvars, eqns, jaxpr.effects)
   if consts is not None:
     return ClosedJaxpr(new_jaxpr, consts)
   return new_jaxpr
@@ -2172,15 +2250,20 @@ def _check_jaxpr(
       if any(isinstance(ina, ConcreteArray) for ina in in_avals):
         raise JaxprTypeError("Equation given ConcreteArray type inputs")
       if prim in custom_typechecks:
-        out_avals = custom_typechecks[prim](*in_avals, **eqn.params)
+        out_avals, effects = custom_typechecks[prim](*in_avals, **eqn.params)
         if out_avals is None:
           out_avals = [v.aval for v in eqn.outvars]
       elif prim.call_primitive:
-        out_avals = check_call(ctx_factory, prim, in_avals, eqn.params)
+        out_avals, effects = check_call(ctx_factory, prim, in_avals, eqn.params)
       elif prim.map_primitive:
-        out_avals = check_map(ctx_factory, prim, in_avals, eqn.params)
+        out_avals, effects = check_map(ctx_factory, prim, in_avals, eqn.params)
       else:
-        out_avals = check_eqn(prim, in_avals, eqn.params)
+        out_avals, effects = check_eqn(prim, in_avals, eqn.params)
+      if eqn.effects != effects:
+        print(eqn.effects, effects)
+        raise JaxprTypeError("Inferred effects do not match equation effects.")
+      if not eqn.effects.issubset(jaxpr.effects):
+        raise JaxprTypeError("Equation effects are not subset of Jaxpr effects.")
       map(write, eqn.outvars, out_avals)
     except JaxprTypeError as e:
       ctx, settings = ctx_factory()
@@ -2196,10 +2279,10 @@ def check_eqn(prim, in_avals, params):
   for jaxpr in jaxprs_in_params(params):
     check_jaxpr(jaxpr)
 
-  out_avals = prim.abstract_eval(*in_avals, **params)
+  out_avals, effects = prim.abstract_eval(*in_avals, **params)
   if not prim.multiple_results:
     out_avals = [out_avals]
-  return out_avals
+  return out_avals, effects
 
 def check_call(ctx_factory, prim, in_avals, params):
   if "call_jaxpr" not in params:
@@ -2221,12 +2304,14 @@ def check_call(ctx_factory, prim, in_avals, params):
   _check_jaxpr(ctx_factory, call_jaxpr, in_avals)
 
   out_avals = [v.aval for v in call_jaxpr.outvars]
-  return out_avals
+  return out_avals, call_jaxpr.effects
 
 def check_map(ctx_factory, prim, in_avals, params):
   if "call_jaxpr" not in params:
     raise JaxprTypeError(f"Map primitive {prim} missing 'call_jaxpr' parameter")
   call_jaxpr = params["call_jaxpr"]
+  if call_jaxpr.effects:
+    raise JaxprTypeError(f"Map primitive {prim} mapping an effectful function")
   if "axis_size" not in params:
     raise JaxprTypeError(f"Map primitive {prim} missing 'axis_size' parameter")
   axis_size = params["axis_size"]
@@ -2257,7 +2342,7 @@ def check_map(ctx_factory, prim, in_avals, params):
   mapped_out_avals = [v.aval for v in call_jaxpr.outvars]
   out_avals = [unmapped_aval(axis_size, axis_name, out_axis, aval) if out_axis is not None else aval
                for aval, out_axis in zip(mapped_out_avals, out_axes)]
-  return out_avals
+  return out_avals, call_jaxpr.effects
 
 
 # ------------------- Jaxpr printed representation -------------------
