@@ -1257,6 +1257,14 @@ _POLY_SHAPE_TEST_HARNESSES = [
                   lambda x: lax.collapse(x, 1, 4),
                   [RandArg((3, 4, 5, 6, 7), _f32)],
                   poly_axes=[(0, 1, 3)]),
+    _make_harness("concatenate", "",
+                  lambda x: jnp.concatenate([x, x], axis=0),
+                  [RandArg((3, 4, 5), _f32)],
+                  poly_axes=[(0, 1)]),
+    _make_harness("concatenate", "grad",
+                  jax.grad(lambda x: jnp.sum(jnp.concatenate([x, x], axis=0))),
+                  [RandArg((3, 4, 5), _f32)],
+                  poly_axes=[(0, 1)]),
     _make_harness("conv_general_dilated", "",
                   lambda lhs, rhs: lax.conv_general_dilated(lhs, rhs,
                                                             window_strides=(2, 3),
@@ -1564,6 +1572,11 @@ _POLY_SHAPE_TEST_HARNESSES = [
                   lambda x: jnp.reshape(x, [2, -1]),
                   [RandArg((3, 4, 5, 6, 7), _f32)],
                   poly_axes=[(0, 2)]),
+    _make_harness("reshape", "_issue_9975",
+                  # The newshape is a scalar
+                  lambda x: jnp.reshape(x, x.shape[0] * x.shape[1]),
+                  [RandArg((3, 4), _f32)],
+                  poly_axes=[0]),
     _make_harness("reshape", "error",
                   lambda x: x.reshape([x.shape[0], -1, 3]),
                   [RandArg((3, 2, 4), _f32)],
