@@ -122,6 +122,11 @@ from jax._src.numpy.lax_numpy import (
     floating as floating,
     fmax as fmax,
     fmin as fmin,
+    frombuffer as frombuffer,
+    fromfile as fromfile,
+    fromfunction as fromfunction,
+    fromiter as fromiter,
+    fromstring as fromstring,
     full as full,
     full_like as full_like,
     gcd as gcd,
@@ -142,7 +147,6 @@ from jax._src.numpy.lax_numpy import (
     iinfo as iinfo,
     indices as indices,
     inexact as inexact,
-    in1d as in1d,
     inf as inf,
     inner as inner,
     insert as insert,
@@ -153,11 +157,9 @@ from jax._src.numpy.lax_numpy import (
     int_ as int_,
     integer as integer,
     interp as interp,
-    intersect1d as intersect1d,
     isclose as isclose,
     iscomplex as iscomplex,
     iscomplexobj as iscomplexobj,
-    isin as isin,
     isreal as isreal,
     isrealobj as isrealobj,
     isscalar as isscalar,
@@ -234,8 +236,6 @@ from jax._src.numpy.lax_numpy import (
     searchsorted as searchsorted,
     select as select,
     set_printoptions as set_printoptions,
-    setdiff1d as setdiff1d,
-    setxor1d as setxor1d,
     shape as shape,
     signedinteger as signedinteger,
     single as single,
@@ -270,8 +270,6 @@ from jax._src.numpy.lax_numpy import (
     uint32 as uint32,
     uint64 as uint64,
     uint8 as uint8,
-    unique as unique,
-    union1d as union1d,
     unpackbits as unpackbits,
     unravel_index as unravel_index,
     unsignedinteger as unsignedinteger,
@@ -308,6 +306,48 @@ from jax._src.numpy.polynomial import (
     roots as roots,
 )
 
+from jax._src.numpy.reductions import (
+    alltrue as alltrue,
+    amin as amin,
+    amax as amax,
+    any as any,
+    all as all,
+    average as average,
+    count_nonzero as count_nonzero,
+    cumsum as cumsum,
+    cumprod as cumprod,
+    cumproduct as cumproduct,
+    max as max,
+    mean as mean,
+    min as min,
+    nancumsum as nancumsum,
+    nancumprod as nancumprod,
+    nanmax as nanmax,
+    nanmean as nanmean,
+    nanmin as nanmin,
+    nanprod as nanprod,
+    nanstd as nanstd,
+    nansum as nansum,
+    nanvar as nanvar,
+    prod as prod, 
+    product as product,
+    ptp as ptp, 
+    sometrue as sometrue, 
+    std as std, 
+    sum as sum, 
+    var as var,
+)
+
+from jax._src.numpy.setops import (
+  in1d as in1d,
+  intersect1d as intersect1d,
+  isin as isin,
+  setdiff1d as setdiff1d,
+  setxor1d as setxor1d,
+  union1d as union1d,
+  unique as unique,
+)
+
 from jax._src.numpy.ufuncs import (
     abs as abs,
     absolute as absolute,
@@ -401,9 +441,6 @@ from jax._src.numpy.ufuncs import (
 
 from jax._src.numpy.vectorize import vectorize as vectorize
 
-# TODO(phawkins): remove this import after fixing users.
-from jax._src.numpy import lax_numpy
-
 # Module initialization is encapsulated in a function to avoid accidental
 # namespace pollution.
 def _init():
